// -*- tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=8 sw=2 sts=2:
#ifndef DUNE_PDELAB_LOCALOPERATOR_DEFAULTIMP_HH
#define DUNE_PDELAB_LOCALOPERATOR_DEFAULTIMP_HH

#include <dune/pdelab/localoperator/numericaljacobian.hh>
#include <dune/pdelab/localoperator/numericaljacobianapply.hh>
#include <dune/pdelab/localoperator/numericalnonlinearjacobianapply.hh>
#include <dune/pdelab/localoperator/numericalresidual.hh>

#endif // DUNE_PDELAB_LOCALOPERATOR_DEFAULTIMP_HH
